from __future__ import print_function, unicode_literals
#import io, json
from twisted.trial import unittest
from ..database import create_channel_db, create_usage_db
from ..server import make_server, CrowdedError

class _Make:
    def make(self, blur_usage=None, with_usage_db=True):
        self._cdb = create_channel_db(":memory:")
        db = create_usage_db(":memory:") if with_usage_db else None
        s = make_server(self._cdb, usage_db=db, blur_usage=blur_usage)
        app = s.get_app("appid")
        return s, db, app

class Current(_Make, unittest.TestCase):
    def test_current_no_mailboxes(self):
        s, db, app = self.make()
        s.dump_stats(456, rebooted=451)
        self.assertEqual(db.execute("SELECT * FROM `current`").fetchall(),
                         [dict(rebooted=451, updated=456, blur_time=None,
                               connections_websocket=0),
                          ])

    def test_current_no_listeners(self):
        s, db, app = self.make()
        app.open_mailbox("m1", "s1", 1)
        s.dump_stats(456, rebooted=451)
        self.assertEqual(db.execute("SELECT * FROM `current`").fetchall(),
                         [dict(rebooted=451, updated=456, blur_time=None,
                               connections_websocket=0),
                          ])

    def test_current_one_listener(self):
        s, db, app = self.make()
        mbox = app.open_mailbox("m1", "s1", 1)
        mbox.add_listener("h1", lambda sm: None, lambda: None)
        s.dump_stats(456, rebooted=451)
        self.assertEqual(db.execute("SELECT * FROM `current`").fetchall(),
                         [dict(rebooted=451, updated=456, blur_time=None,
                               connections_websocket=1),
                          ])

class ClientVersion(_Make, unittest.TestCase):
    def test_add_version(self):
        s, db, app = self.make()
        app.log_client_version(451, "side1", ("python", "1.2.3"))
        self.assertEqual(db.execute("SELECT * FROM `client_versions`").fetchall(),
                         [dict(app_id="appid", connect_time=451, side="side1",
                               implementation="python", version="1.2.3")])

    def test_add_version_extra_fields(self):
        s, db, app = self.make()
        app.log_client_version(451, "side1", ("python", "1.2.3", "extra"))
        self.assertEqual(db.execute("SELECT * FROM `client_versions`").fetchall(),
                         [dict(app_id="appid", connect_time=451, side="side1",
                               implementation="python", version="1.2.3")])

    def test_blur(self):
        s, db, app = self.make(blur_usage=100)
        app.log_client_version(451, "side1", ("python", "1.2.3"))
        self.assertEqual(db.execute("SELECT * FROM `client_versions`").fetchall(),
                         [dict(app_id="appid", connect_time=400, side="side1",
                               implementation="python", version="1.2.3")])

    def test_no_usage_db(self):
        s, db, app = self.make(with_usage_db=False)
        app.log_client_version(451, "side1", ("python", "1.2.3"))

class Nameplate(_Make, unittest.TestCase):
    def test_nameplate_happy(self):
        s, db, app = self.make()
        app.claim_nameplate("n1", "s1", 1)
        app.claim_nameplate("n1", "s2", 3)
        app.release_nameplate("n1", "s1", 6)
        self.assertEqual(db.execute("SELECT * FROM `nameplates`").fetchall(),
                         [])
        app.release_nameplate("n1", "s2", 10)
        self.assertEqual(db.execute("SELECT * FROM `nameplates`").fetchall(),
                         [dict(app_id="appid", result="happy",
                               started=1, waiting_time=2, total_time=9)])

    def test_nameplate_lonely(self):
        s, db, app = self.make()
        app.claim_nameplate("n1", "s1", 1)
        app.release_nameplate("n1", "s1", 6)
        self.assertEqual(db.execute("SELECT * FROM `nameplates`").fetchall(),
                         [dict(app_id="appid", result="lonely",
                               started=1, waiting_time=None, total_time=5)])

    def test_nameplate_pruney(self):
        s, db, app = self.make()
        app.claim_nameplate("n1", "s1", 1)
        app.prune(10, 5) # prune at t=10, anything earlier than 5 is "old"
        self.assertEqual(db.execute("SELECT * FROM `nameplates`").fetchall(),
                         [dict(app_id="appid", result="pruney",
                               started=1, waiting_time=None, total_time=9)])

    def test_nameplate_crowded(self):
        s, db, app = self.make()
        app.claim_nameplate("n1", "s1", 1)
        app.claim_nameplate("n1", "s2", 2)
        with self.assertRaises(CrowdedError):
            app.claim_nameplate("n1", "s3", 3)
        self.assertEqual(db.execute("SELECT * FROM `nameplates`").fetchall(),
                         [])
        app.release_nameplate("n1", "s1", 4)
        self.assertEqual(db.execute("SELECT * FROM `nameplates`").fetchall(),
                         [])
        app.release_nameplate("n1", "s2", 5)
        self.assertEqual(db.execute("SELECT * FROM `nameplates`").fetchall(),
                         [])
        #print(self._cdb.execute("SELECT * FROM `nameplates`").fetchall())
        #print(self._cdb.execute("SELECT * FROM `nameplate_sides`").fetchall())
        # TODO: to get "crowded", we need all three sides to release the
        # nameplate, even though the third side threw CrowdedError and thus
        # probably doesn't think it has a claim
        app.release_nameplate("n1", "s3", 6)
        self.assertEqual(db.execute("SELECT * FROM `nameplates`").fetchall(),
                         [dict(app_id="appid", result="crowded",
                               started=1, waiting_time=1, total_time=5)])

    def test_nameplate_crowded_pruned(self):
        s, db, app = self.make()
        app.claim_nameplate("n1", "s1", 1)
        app.claim_nameplate("n1", "s2", 2)
        with self.assertRaises(CrowdedError):
            app.claim_nameplate("n1", "s3", 3)
        self.assertEqual(db.execute("SELECT * FROM `nameplates`").fetchall(),
                         [])
        app.prune(10, 5)
        self.assertEqual(db.execute("SELECT * FROM `nameplates`").fetchall(),
                         [dict(app_id="appid", result="crowded",
                               started=1, waiting_time=1, total_time=9)])

    def test_no_db(self):
        s, db, app = self.make(with_usage_db=False)
        app.claim_nameplate("n1", "s1", 1)
        app.release_nameplate("n1", "s1", 6)
        s.dump_stats(3, 1)

    def test_nameplate_happy_blur_usage(self):
        s, db, app = self.make(blur_usage=20)
        app.claim_nameplate("n1", "s1", 21)
        app.claim_nameplate("n1", "s2", 23)
        app.release_nameplate("n1", "s1", 26)
        self.assertEqual(db.execute("SELECT * FROM `nameplates`").fetchall(),
                         [])
        app.release_nameplate("n1", "s2", 30)
        self.assertEqual(db.execute("SELECT * FROM `nameplates`").fetchall(),
                         [dict(app_id="appid", result="happy",
                               started=20, waiting_time=2, total_time=9)])

class Mailbox(_Make, unittest.TestCase):
    def test_mailbox_prune_quiet(self):
        s, db, app = self.make()
        app.claim_nameplate("n1", "s1", 1)
        app.release_nameplate("n1", "s1", 2)
        app.prune(10, 5)
        self.assertEqual(db.execute("SELECT * FROM `mailboxes`").fetchall(),
                         [dict(app_id="appid", for_nameplate=1, result="pruney",
                               started=1, waiting_time=None, total_time=9)])

    def test_mailbox_lonely(self):
        s, db, app = self.make()
        mid = app.claim_nameplate("n1", "s1", 1)
        mbox = app.open_mailbox(mid, "s1", 2)
        app.release_nameplate("n1", "s1", 3)
        mbox.close("s1", "mood-ignored", 4)
        self.assertEqual(db.execute("SELECT * FROM `mailboxes`").fetchall(),
                         [dict(app_id="appid", for_nameplate=1, result="lonely",
                               started=1, waiting_time=None, total_time=3)])

    def test_mailbox_happy(self):
        s, db, app = self.make()
        mid = app.claim_nameplate("n1", "s1", 1)
        mbox1 = app.open_mailbox(mid, "s1", 2)
        app.release_nameplate("n1", "s1", 3)
        mbox2 = app.open_mailbox(mid, "s2", 4)
        mbox1.close("s1", "happy", 5)
        mbox2.close("s2", "happy", 6)
        self.assertEqual(db.execute("SELECT * FROM `mailboxes`").fetchall(),
                         [dict(app_id="appid", for_nameplate=1, result="happy",
                               started=1, waiting_time=3, total_time=5)])

    def test_mailbox_happy_blur_usage(self):
        s, db, app = self.make(blur_usage=20)
        mid = app.claim_nameplate("n1", "s1", 21)
        mbox1 = app.open_mailbox(mid, "s1", 22)
        app.release_nameplate("n1", "s1", 23)
        mbox2 = app.open_mailbox(mid, "s2", 24)
        mbox1.close("s1", "happy", 25)
        mbox2.close("s2", "happy", 26)
        self.assertEqual(db.execute("SELECT * FROM `mailboxes`").fetchall(),
                         [dict(app_id="appid", for_nameplate=1, result="happy",
                               started=20, waiting_time=3, total_time=5)])

    def test_mailbox_lonely_connected(self):
        # I don't think this could actually happen. It requires both sides to
        # connect, but then at least one side says they're lonely when they
        # close.
        s, db, app = self.make()
        mid = app.claim_nameplate("n1", "s1", 1)
        mbox1 = app.open_mailbox(mid, "s1", 2)
        app.release_nameplate("n1", "s1", 3)
        mbox2 = app.open_mailbox(mid, "s2", 4)
        mbox1.close("s1", "lonely", 5)
        mbox2.close("s2", "happy", 6)
        self.assertEqual(db.execute("SELECT * FROM `mailboxes`").fetchall(),
                         [dict(app_id="appid", for_nameplate=1, result="lonely",
                               started=1, waiting_time=3, total_time=5)])

    def test_mailbox_scary(self):
        s, db, app = self.make()
        mid = app.claim_nameplate("n1", "s1", 1)
        mbox1 = app.open_mailbox(mid, "s1", 2)
        app.release_nameplate("n1", "s1", 3)
        mbox2 = app.open_mailbox(mid, "s2", 4)
        mbox1.close("s1", "scary", 5)
        mbox2.close("s2", "happy", 6)
        self.assertEqual(db.execute("SELECT * FROM `mailboxes`").fetchall(),
                         [dict(app_id="appid", for_nameplate=1, result="scary",
                               started=1, waiting_time=3, total_time=5)])

    def test_mailbox_errory(self):
        s, db, app = self.make()
        mid = app.claim_nameplate("n1", "s1", 1)
        mbox1 = app.open_mailbox(mid, "s1", 2)
        app.release_nameplate("n1", "s1", 3)
        mbox2 = app.open_mailbox(mid, "s2", 4)
        mbox1.close("s1", "errory", 5)
        mbox2.close("s2", "happy", 6)
        self.assertEqual(db.execute("SELECT * FROM `mailboxes`").fetchall(),
                         [dict(app_id="appid", for_nameplate=1, result="errory",
                               started=1, waiting_time=3, total_time=5)])

    def test_mailbox_errory_scary(self):
        s, db, app = self.make()
        mid = app.claim_nameplate("n1", "s1", 1)
        mbox1 = app.open_mailbox(mid, "s1", 2)
        app.release_nameplate("n1", "s1", 3)
        mbox2 = app.open_mailbox(mid, "s2", 4)
        mbox1.close("s1", "errory", 5)
        mbox2.close("s2", "scary", 6)
        self.assertEqual(db.execute("SELECT * FROM `mailboxes`").fetchall(),
                         [dict(app_id="appid", for_nameplate=1, result="scary",
                               started=1, waiting_time=3, total_time=5)])

    def test_mailbox_crowded(self):
        s, db, app = self.make()
        mid = app.claim_nameplate("n1", "s1", 1)
        mbox1 = app.open_mailbox(mid, "s1", 2)
        app.release_nameplate("n1", "s1", 3)
        mbox2 = app.open_mailbox(mid, "s2", 4)
        with self.assertRaises(CrowdedError):
            app.open_mailbox(mid, "s3", 5)
        mbox1.close("s1", "happy", 6)
        mbox2.close("s2", "happy", 7)
        # again, not realistic
        mbox2.close("s3", "happy", 8)
        self.assertEqual(db.execute("SELECT * FROM `mailboxes`").fetchall(),
                         [dict(app_id="appid", for_nameplate=1, result="crowded",
                               started=1, waiting_time=3, total_time=7)])

## class LogToStdout(unittest.TestCase):
##     def test_log(self):
##         # emit lines of JSON to log_file, if set
##         log_file = io.StringIO()
##         t = Transit(blur_usage=None, log_file=log_file, usage_db=None)
##         t.recordUsage(started=123, result="happy", total_bytes=100,
##                       total_time=10, waiting_time=2)
##         self.assertEqual(json.loads(log_file.getvalue()),
##                          {"started": 123, "total_time": 10,
##                           "waiting_time": 2, "total_bytes": 100,
##                           "mood": "happy"})

##     def test_log_blurred(self):
##         # if blurring is enabled, timestamps should be rounded to the
##         # requested amount, and sizes should be rounded up too
##         log_file = io.StringIO()
##         t = Transit(blur_usage=60, log_file=log_file, usage_db=None)
##         t.recordUsage(started=123, result="happy", total_bytes=11999,
##                       total_time=10, waiting_time=2)
##         self.assertEqual(json.loads(log_file.getvalue()),
##                          {"started": 120, "total_time": 10,
##                           "waiting_time": 2, "total_bytes": 20000,
##                           "mood": "happy"})

##     def test_do_not_log(self):
##         t = Transit(blur_usage=60, log_file=None, usage_db=None)
##         t.recordUsage(started=123, result="happy", total_bytes=11999,
##                       total_time=10, waiting_time=2)



import os, json, time
from twisted.internet import reactor
from twisted.python import usage, log
from twisted.application.service import MultiService
from twisted.application.internet import (TimerService,
                                          StreamServerEndpointService)
from twisted.internet import endpoints
from .increase_rlimits import increase_rlimits
from .server import make_server
from .web import make_web_server
from .database import create_or_upgrade_channel_db, create_or_upgrade_usage_db

LONGDESC = """This plugin sets up a 'Mailbox' server for magic-wormhole.
This service forwards short messages between clients, to perform key exchange
and connection setup."""

class Options(usage.Options):
    synopsis = "[--port=] [--log-fd] [--blur-usage=] [--usage-db=]"
    longdesc = LONGDESC

    optParameters = [
        ("port", "p", "tcp:4000:interface=\:\:", "endpoint to listen on"),
        ("blur-usage", None, None, "round logged access times to improve privacy"),
        ("log-fd", None, None, "write JSON usage logs to this file descriptor"),
        ("channel-db", None, "relay.sqlite", "location for the state database"),
        ("usage-db", None, None, "record usage data (SQLite)"),
        ("advertise-version", None, None, "version to recommend to clients"),
        ("signal-error", None, None, "force all clients to fail with a message"),
        ("motd", None, None, "Send a Message of the Day in the welcome"),
        ]
    optFlags = [
        ("disallow-list", None, "refuse to send list of allocated nameplates"),
        ]

    def __init__(self):
        super(Options, self).__init__()
        self["websocket-protocol-options"] = []
        self["allow-list"] = True

    def opt_disallow_list(self):
        self["allow-list"] = False

    def opt_log_fd(self, arg):
        self["log-fd"] = int(arg)

    def opt_blur_usage(self, arg):
        # --blur-usage= is in seconds. If the option isn't provided, we'll keep
        # the default of None
        self["blur-usage"] = int(arg)

    def opt_websocket_protocol_option(self, arg):
        """A websocket server protocol option to configure: OPTION=VALUE. This option can be provided multiple times."""
        try:
            key, value = arg.split("=", 1)
        except ValueError:
            raise usage.UsageError("format options as OPTION=VALUE")
        try:
            value = json.loads(value)
        except:
            raise usage.UsageError("could not parse JSON value for {}".format(key))
        self["websocket-protocol-options"].append((key, value))


SECONDS = 1.0
MINUTE = 60*SECONDS

# CHANNEL_EXPIRATION_TIME should be longer than EXPIRATION_CHECK_PERIOD
CHANNEL_EXPIRATION_TIME = 11*MINUTE
EXPIRATION_CHECK_PERIOD = 5*MINUTE

def makeService(config, channel_db="relay.sqlite", reactor=reactor):
    increase_rlimits()

    parent = MultiService()

    channel_db = create_or_upgrade_channel_db(config["channel-db"])
    usage_db = create_or_upgrade_usage_db(config["usage-db"])
    log_file = (os.fdopen(int(config["log-fd"]), "w")
                if config["log-fd"] is not None
                else None)
    server = make_server(channel_db,
                         allow_list=config["allow-list"],
                         advertise_version=config["advertise-version"],
                         signal_error=config["signal-error"],
                         blur_usage=config["blur-usage"],
                         usage_db=usage_db,
                         log_file=log_file,
                         welcome_motd=config["motd"],
                         )
    server.setServiceParent(parent)
    rebooted = time.time()
    def expire():
        now = time.time()
        old = now - CHANNEL_EXPIRATION_TIME
        try:
            server.prune_all_apps(now, old)
        except Exception as e:
            # catch-and-log exceptions during prune, so a single error won't
            # kill the loop. See #13 for details.
            log.msg("error during prune_all_apps")
            log.err(e)
        server.dump_stats(now, rebooted=rebooted)
    TimerService(EXPIRATION_CHECK_PERIOD, expire).setServiceParent(parent)

    log_requests = config["blur-usage"] is None
    site = make_web_server(server, log_requests,
                           config["websocket-protocol-options"])
    ep = endpoints.serverFromString(reactor, config["port"]) # to listen
    StreamServerEndpointService(ep, site).setServiceParent(parent)
    log.msg("websocket listening on ws://HOSTNAME:PORT/v1")

    return parent



from __future__ import unicode_literals
import os, shutil
import sqlite3
import tempfile
from pkg_resources import resource_string
from twisted.python import log

class DBError(Exception):
    pass

def get_schema(name, version):
    schema_bytes = resource_string("wormhole_mailbox_server",
                                   "db-schemas/%s-v%d.sql" % (name, version))
    return schema_bytes.decode("utf-8")

def get_upgrader(name, new_version):
    try:
        schema_bytes = resource_string("wormhole_mailbox_server",
                                       "db-schemas/upgrade-%s-to-v%d.sql" %
                                       (name, new_version))
    except EnvironmentError: # includes FileNotFoundError on py3
        raise ValueError("no upgrader for %d" % new_version)
    return schema_bytes.decode("utf-8")

CHANNELDB_TARGET_VERSION = 1
USAGEDB_TARGET_VERSION = 2

def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d

def _initialize_db_schema(db, name, target_version):
    """Creates the application schema in the given database.
    """
    log.msg("populating new database with schema %s v%s" % (name, target_version))
    schema = get_schema(name, target_version)
    db.executescript(schema)
    db.execute("INSERT INTO version (version) VALUES (?)",
               (target_version,))
    db.commit()

def _initialize_db_connection(db):
    """Sets up the db connection object with a row factory and with necessary
    foreign key settings.
    """
    db.row_factory = dict_factory
    db.execute("PRAGMA foreign_keys = ON")
    problems = db.execute("PRAGMA foreign_key_check").fetchall()
    if problems:
        raise DBError("failed foreign key check: %s" % (problems,))

def _open_db_connection(dbfile):
    """Open a new connection to the SQLite3 database at the given path.
    """
    try:
        db = sqlite3.connect(dbfile)
        _initialize_db_connection(db)
    except (EnvironmentError, sqlite3.OperationalError, sqlite3.DatabaseError) as e:
        # this indicates that the file is not a compatible database format.
        # Perhaps it was created with an old version, or it might be junk.
        raise DBError("Unable to create/open db file %s: %s" % (dbfile, e))
    return db

def _get_temporary_dbfile(dbfile):
    """Get a temporary filename near the given path.
    """
    fd, name = tempfile.mkstemp(
        prefix=os.path.basename(dbfile) + ".",
        dir=os.path.dirname(dbfile)
    )
    os.close(fd)
    return name

def _atomic_create_and_initialize_db(dbfile, name, target_version):
    """Create and return a new database, initialized with the application
    schema.

    If anything goes wrong, nothing is left at the ``dbfile`` path.
    """
    temp_dbfile = _get_temporary_dbfile(dbfile)
    db = _open_db_connection(temp_dbfile)
    _initialize_db_schema(db, name, target_version)
    db.close()
    os.rename(temp_dbfile, dbfile)
    return _open_db_connection(dbfile)

def _get_db(dbfile, name, target_version):
    """Open or create the given db file. The parent directory must exist.
    Returns the db connection object, or raises DBError.
    """
    if dbfile == ":memory:":
        db = _open_db_connection(dbfile)
        _initialize_db_schema(db, name, target_version)
    elif os.path.exists(dbfile):
        db = _open_db_connection(dbfile)
    else:
        db = _atomic_create_and_initialize_db(dbfile, name, target_version)

    version = db.execute("SELECT version FROM version").fetchone()["version"]

    if version < target_version and dbfile != ":memory:":
        backup_fn = "%s-backup-v%d" % (dbfile, version)
        log.msg(" storing backup of v%d db in %s" % (version, backup_fn))
        shutil.copy(dbfile, backup_fn)

    while version < target_version:
        log.msg(" need to upgrade from %s to %s" % (version, target_version))
        try:
            upgrader = get_upgrader(name, version+1)
        except ValueError:
            log.msg(" unable to upgrade %s to %s" % (version, version+1))
            raise DBError("Unable to upgrade %s to version %s, left at %s"
                          % (dbfile, version+1, version))
        log.msg(" executing upgrader v%s->v%s" % (version, version+1))
        db.executescript(upgrader)
        db.commit()
        version = version+1

    if version != target_version:
        raise DBError("Unable to handle db version %s" % version)

    return db

def create_or_upgrade_channel_db(dbfile):
    return _get_db(dbfile, "channel", CHANNELDB_TARGET_VERSION)

def create_or_upgrade_usage_db(dbfile):
    if dbfile is None:
        return None
    return _get_db(dbfile, "usage", USAGEDB_TARGET_VERSION)

class DBDoesntExist(Exception):
    pass

def open_existing_db(dbfile):
    assert dbfile != ":memory:"
    if not os.path.exists(dbfile):
        raise DBDoesntExist()
    return _open_db_connection(dbfile)

class DBAlreadyExists(Exception):
    pass

def create_channel_db(dbfile):
    """Create the given db file. Refuse to touch a pre-existing file.

    This is meant for use by migration tools, to create the output target"""

    if dbfile == ":memory:":
        db = _open_db_connection(dbfile)
        _initialize_db_schema(db, "channel", CHANNELDB_TARGET_VERSION)
    elif os.path.exists(dbfile):
        raise DBAlreadyExists()
    else:
        db = _atomic_create_and_initialize_db(dbfile, "channel",
                                              CHANNELDB_TARGET_VERSION)
    return db

def create_usage_db(dbfile):
    if dbfile == ":memory:":
        db = _open_db_connection(dbfile)
        _initialize_db_schema(db, "usage", USAGEDB_TARGET_VERSION)
    elif os.path.exists(dbfile):
        raise DBAlreadyExists()
    else:
        db = _atomic_create_and_initialize_db(dbfile, "usage",
                                              USAGEDB_TARGET_VERSION)
    return db

def dump_db(db):
    # to let _iterdump work, we need to restore the original row factory
    orig = db.row_factory
    try:
        db.row_factory = sqlite3.Row
        return "".join(db.iterdump())
    finally:
        db.row_factory = orig



from __future__ import print_function, unicode_literals
import io, time
from unittest import mock
import treq
from twisted.trial import unittest
from twisted.internet import defer, reactor
from twisted.internet.defer import inlineCallbacks, returnValue
from ..web import make_web_server
from ..server import SidedMessage
from ..database import create_or_upgrade_usage_db
from .common import ServerBase, _Util
from .ws_client import WSFactory

class WebSocketProtocolOptions(unittest.TestCase):
    @mock.patch('wormhole_mailbox_server.web.WebSocketServerFactory')
    def test_set(self, fake_factory):
        make_web_server(None, False,
                        websocket_protocol_options=[ ("foo", "bar"), ],
                        )
        self.assertEqual(
            mock.call().setProtocolOptions(foo="bar"),
            fake_factory.mock_calls[1],
        )

class LogRequests(ServerBase, unittest.TestCase):
    def setUp(self):
        self._clients = []

    def tearDown(self):
        for c in self._clients:
            c.transport.loseConnection()
        return ServerBase.tearDown(self)

    @inlineCallbacks
    def make_client(self):
        f = WSFactory(self.relayurl)
        f.d = defer.Deferred()
        reactor.connectTCP("127.0.0.1", self.rdv_ws_port, f)
        c = yield f.d
        self._clients.append(c)
        returnValue(c)

    @inlineCallbacks
    def test_log_http(self):
        yield self._setup_relay(do_listen=True, web_log_requests=True)
        # check the HTTP log
        fakelog = io.BytesIO()
        self._site.logFile = fakelog
        yield treq.get("http://127.0.0.1:%d/" % self.rdv_ws_port,
                       persistent=False)
        lines = fakelog.getvalue().splitlines()
        self.assertEqual(len(lines), 1, lines)

    @inlineCallbacks
    def test_log_websocket(self):
        yield self._setup_relay(do_listen=True, web_log_requests=True)
        # now check the twisted log for websocket connect messages
        with mock.patch("wormhole_mailbox_server.server_websocket.log.msg") as l:
            c1 = yield self.make_client()
            yield c1.next_non_ack()
            # the actual message includes the TCP port number of the client
            client_port = self._clients[0].transport.getHost().port
            expected = "ws client connecting: tcp4:127.0.0.1:%d" % client_port
            self.assertEqual(l.mock_calls, [mock.call(expected)])


    @inlineCallbacks
    def test_no_log_http(self):
        yield self._setup_relay(do_listen=True, web_log_requests=False)
        # check the HTTP log
        fakelog = io.BytesIO()
        self._site.logFile = fakelog
        yield treq.get("http://127.0.0.1:%d/" % self.rdv_ws_port,
                       persistent=False)
        lines = fakelog.getvalue().splitlines()
        self.assertEqual(len(lines), 0, lines)

    @inlineCallbacks
    def test_no_log_websocket(self):
        yield self._setup_relay(do_listen=True,
                                blur_usage=60, web_log_requests=True)
        # now check the twisted log for websocket connect messages
        with mock.patch("wormhole_mailbox_server.server_websocket.log.msg") as l:
            c1 = yield self.make_client()
            yield c1.next_non_ack()
            self.assertEqual(l.mock_calls, [])


class WebSocketAPI(_Util, ServerBase, unittest.TestCase):
    @inlineCallbacks
    def setUp(self):
        self._lp = None
        self._clients = []
        self._usage_db = usage_db = create_or_upgrade_usage_db(":memory:")
        yield self._setup_relay(do_listen=True,
                                advertise_version="advertised.version",
                                usage_db=usage_db)

    def tearDown(self):
        for c in self._clients:
            c.transport.loseConnection()
        return ServerBase.tearDown(self)

    @inlineCallbacks
    def make_client(self):
        f = WSFactory(self.relayurl)
        f.d = defer.Deferred()
        reactor.connectTCP("127.0.0.1", self.rdv_ws_port, f)
        c = yield f.d
        self._clients.append(c)
        returnValue(c)

    def check_welcome(self, data):
        self.failUnlessIn("welcome", data)
        self.failUnlessEqual(data["welcome"],
                             {"current_cli_version": "advertised.version"})

    @inlineCallbacks
    def test_welcome(self):
        c1 = yield self.make_client()
        msg = yield c1.next_non_ack()
        self.check_welcome(msg)
        self.assertEqual(self._server._apps, {})

    @inlineCallbacks
    def test_bind(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()

        c1.send("bind", appid="appid") # missing side=
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "bind requires 'side'")

        c1.send("bind", side="side") # missing appid=
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "bind requires 'appid'")

        c1.send("bind", appid="appid", side="side")
        yield c1.sync()
        self.assertEqual(list(self._server._apps.keys()), ["appid"])

        c1.send("bind", appid="appid", side="side") # duplicate
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "already bound")

        c1.send_notype(other="misc") # missing 'type'
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "missing 'type'")

        c1.send("___unknown") # unknown type
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "unknown type")

        c1.send("ping") # missing 'ping'
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "ping requires 'ping'")

    @inlineCallbacks
    def test_bind_with_client_version(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()

        c1.send("bind", appid="appid", side="side",
                client_version=("python", "1.2.3"))
        yield c1.sync()
        self.assertEqual(list(self._server._apps.keys()), ["appid"])
        v = self._usage_db.execute("SELECT * FROM `client_versions`").fetchall()
        self.assertEqual(v[0]["app_id"], "appid")
        self.assertEqual(v[0]["side"], "side")
        self.assertEqual(v[0]["implementation"], "python")
        self.assertEqual(v[0]["version"], "1.2.3")

    @inlineCallbacks
    def test_bind_without_client_version(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()

        c1.send("bind", appid="appid", side="side")
        yield c1.sync()
        self.assertEqual(list(self._server._apps.keys()), ["appid"])
        v = self._usage_db.execute("SELECT * FROM `client_versions`").fetchall()
        self.assertEqual(v[0]["app_id"], "appid")
        self.assertEqual(v[0]["side"], "side")
        self.assertEqual(v[0]["implementation"], None)
        self.assertEqual(v[0]["version"], None)

    @inlineCallbacks
    def test_bind_with_client_version_extra_junk(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()

        c1.send("bind", appid="appid", side="side",
                client_version=("python", "1.2.3", "extra ignore me"))
        yield c1.sync()
        self.assertEqual(list(self._server._apps.keys()), ["appid"])
        v = self._usage_db.execute("SELECT * FROM `client_versions`").fetchall()
        self.assertEqual(v[0]["app_id"], "appid")
        self.assertEqual(v[0]["side"], "side")
        self.assertEqual(v[0]["implementation"], "python")
        self.assertEqual(v[0]["version"], "1.2.3")

    @inlineCallbacks
    def test_list(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()

        c1.send("list") # too early, must bind first
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "must bind first")

        c1.send("bind", appid="appid", side="side")
        c1.send("list")
        m = yield c1.next_non_ack()
        self.assertEqual(m["type"], "nameplates")
        self.assertEqual(m["nameplates"], [])

        app = self._server.get_app("appid")
        nameplate_id1 = app.allocate_nameplate("side", 0)
        app.claim_nameplate("np2", "side", 0)

        c1.send("list")
        m = yield c1.next_non_ack()
        self.assertEqual(m["type"], "nameplates")
        nids = set()
        for n in m["nameplates"]:
            self.assertEqual(type(n), dict)
            self.assertEqual(list(n.keys()), ["id"])
            nids.add(n["id"])
        self.assertEqual(nids, set([nameplate_id1, "np2"]))

    @inlineCallbacks
    def test_allocate(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()

        c1.send("allocate") # too early, must bind first
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "must bind first")

        c1.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")
        c1.send("allocate")
        m = yield c1.next_non_ack()
        self.assertEqual(m["type"], "allocated")
        name = m["nameplate"]

        nids = app.get_nameplate_ids()
        self.assertEqual(len(nids), 1)
        self.assertEqual(name, list(nids)[0])

        c1.send("allocate")
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"],
                         "you already allocated one, don't be greedy")

        c1.send("claim", nameplate=name) # allocate+claim is ok
        yield c1.sync()
        np_row, side_rows = self._nameplate(app, name)
        self.assertEqual(len(side_rows), 1)
        self.assertEqual(side_rows[0]["side"], "side")

    @inlineCallbacks
    def test_claim(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")

        c1.send("claim") # missing nameplate=
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "claim requires 'nameplate'")

        c1.send("claim", nameplate="np1")
        m = yield c1.next_non_ack()
        self.assertEqual(m["type"], "claimed")
        mailbox_id = m["mailbox"]
        self.assertEqual(type(mailbox_id), type(""))

        c1.send("claim", nameplate="np1")
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error", err)
        self.assertEqual(err["error"], "only one claim per connection")

        nids = app.get_nameplate_ids()
        self.assertEqual(len(nids), 1)
        self.assertEqual("np1", list(nids)[0])
        np_row, side_rows = self._nameplate(app, "np1")
        self.assertEqual(len(side_rows), 1)
        self.assertEqual(side_rows[0]["side"], "side")

        # claiming a nameplate assigns a random mailbox id and creates the
        # mailbox row
        mailboxes = app._db.execute("SELECT * FROM `mailboxes`"
                                    " WHERE `app_id`='appid'").fetchall()
        self.assertEqual(len(mailboxes), 1)

    @inlineCallbacks
    def test_claim_crowded(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")

        app.claim_nameplate("np1", "side1", 0)
        app.claim_nameplate("np1", "side2", 0)

        # the third claim will signal crowding
        c1.send("claim", nameplate="np1")
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "crowded")

    @inlineCallbacks
    def test_release(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")

        app.claim_nameplate("np1", "side2", 0)

        c1.send("release") # didn't do claim first
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"],
                         "release without nameplate must follow claim")

        c1.send("claim", nameplate="np1")
        yield c1.next_non_ack()

        c1.send("release")
        m = yield c1.next_non_ack()
        self.assertEqual(m["type"], "released", m)

        np_row, side_rows = self._nameplate(app, "np1")
        claims = [(row["side"], row["claimed"]) for row in side_rows]
        self.assertIn(("side", False), claims)
        self.assertIn(("side2", True), claims)

        c1.send("release") # no longer claimed
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "only one release per connection")

    @inlineCallbacks
    def test_release_named(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")

        c1.send("claim", nameplate="np1")
        yield c1.next_non_ack()

        c1.send("release", nameplate="np1")
        m = yield c1.next_non_ack()
        self.assertEqual(m["type"], "released", m)

    @inlineCallbacks
    def test_release_named_ignored(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")

        c1.send("release", nameplate="np1") # didn't do claim first, ignored
        m = yield c1.next_non_ack()
        self.assertEqual(m["type"], "released", m)

    @inlineCallbacks
    def test_release_named_mismatch(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")

        c1.send("claim", nameplate="np1")
        yield c1.next_non_ack()

        c1.send("release", nameplate="np2") # mismatching nameplate
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"],
                         "release and claim must use same nameplate")

    @inlineCallbacks
    def test_open(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")

        c1.send("open") # missing mailbox=
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "open requires 'mailbox'")

        mb1 = app.open_mailbox("mb1", "side2", 0)
        mb1.add_message(SidedMessage(side="side2", phase="phase",
                                     body="body", server_rx=0,
                                     msg_id="msgid"))

        c1.send("open", mailbox="mb1")
        m = yield c1.next_non_ack()
        self.assertEqual(m["type"], "message")
        self.assertEqual(m["body"], "body")
        self.assertTrue(mb1.has_listeners())

        mb1.add_message(SidedMessage(side="side2", phase="phase2",
                                     body="body2", server_rx=0,
                                     msg_id="msgid"))
        m = yield c1.next_non_ack()
        self.assertEqual(m["type"], "message")
        self.assertEqual(m["body"], "body2")

        c1.send("open", mailbox="mb1")
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "only one open per connection")

        # exercise the _stop() handler too, which is a nop
        mb1.close("side2", "happy", 1)
        mb1.close("side", "happy", 2)

    @inlineCallbacks
    def test_open_crowded(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")

        mbid = app.claim_nameplate("np1", "side1", 0)
        app.claim_nameplate("np1", "side2", 0)

        # the third open will signal crowding
        c1.send("open", mailbox=mbid)
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "crowded")

    @inlineCallbacks
    def test_add(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")
        mb1 = app.open_mailbox("mb1", "side2", 0)
        l1 = []; stop1 = []; stop1_f = lambda: stop1.append(True)
        mb1.add_listener("handle1", l1.append, stop1_f)

        c1.send("add") # didn't open first
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "must open mailbox before adding")

        c1.send("open", mailbox="mb1")

        c1.send("add", body="body") # missing phase=
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "missing 'phase'")

        c1.send("add", phase="phase") # missing body=
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "missing 'body'")

        c1.send("add", phase="phase", body="body")
        m = yield c1.next_non_ack() # echoed back
        self.assertEqual(m["type"], "message")
        self.assertEqual(m["body"], "body")

        self.assertEqual(len(l1), 1)
        self.assertEqual(l1[0].body, "body")

    @inlineCallbacks
    def test_close(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")

        c1.send("close", mood="mood") # must open first
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "close without mailbox must follow open")

        c1.send("open", mailbox="mb1")
        yield c1.sync()
        mb1 = app._mailboxes["mb1"]
        self.assertTrue(mb1.has_listeners())

        c1.send("close", mood="mood")
        m = yield c1.next_non_ack()
        self.assertEqual(m["type"], "closed")
        self.assertFalse(mb1.has_listeners())

        c1.send("close", mood="mood") # already closed
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error", m)
        self.assertEqual(err["error"], "only one close per connection")

    @inlineCallbacks
    def test_close_named(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")

        c1.send("open", mailbox="mb1")
        yield c1.sync()

        c1.send("close", mailbox="mb1", mood="mood")
        m = yield c1.next_non_ack()
        self.assertEqual(m["type"], "closed")

    @inlineCallbacks
    def test_close_named_ignored(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")

        c1.send("close", mailbox="mb1", mood="mood") # no open first, ignored
        m = yield c1.next_non_ack()
        self.assertEqual(m["type"], "closed")

    @inlineCallbacks
    def test_close_named_mismatch(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")

        c1.send("open", mailbox="mb1")
        yield c1.sync()

        c1.send("close", mailbox="mb2", mood="mood")
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "open and close must use same mailbox")

    @inlineCallbacks
    def test_close_crowded(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")

        mbid = app.claim_nameplate("np1", "side1", 0)
        app.claim_nameplate("np1", "side2", 0)

        # a close that allocates a third side will signal crowding
        c1.send("close", mailbox=mbid)
        err = yield c1.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "crowded")


    @inlineCallbacks
    def test_disconnect(self):
        c1 = yield self.make_client()
        yield c1.next_non_ack()
        c1.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")

        c1.send("open", mailbox="mb1")
        yield c1.sync()
        mb1 = app._mailboxes["mb1"]
        self.assertTrue(mb1.has_listeners())

        yield c1.close()
        # wait for the server to notice the socket has closed
        started = time.time()
        while mb1.has_listeners() and (time.time()-started < 5.0):
            d = defer.Deferred()
            reactor.callLater(0.01, d.callback, None)
            yield d
        self.assertFalse(mb1.has_listeners())

    @inlineCallbacks
    def test_interrupted_client_nameplate(self):
        # a client's interactions with the server might be split over
        # multiple sequential WebSocket connections, e.g. when the server is
        # bounced and the client reconnects, or vice versa
        c = yield self.make_client()
        yield c.next_non_ack()
        c.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")

        c.send("claim", nameplate="np1")
        m = yield c.next_non_ack()
        self.assertEqual(m["type"], "claimed")
        mailbox_id = m["mailbox"]
        self.assertEqual(type(mailbox_id), type(""))
        np_row, side_rows = self._nameplate(app, "np1")
        claims = [(row["side"], row["claimed"]) for row in side_rows]
        self.assertEqual(claims, [("side", True)])
        c.close()
        yield c.d

        c = yield self.make_client()
        yield c.next_non_ack()
        c.send("bind", appid="appid", side="side")
        c.send("claim", nameplate="np1") # idempotent
        m = yield c.next_non_ack()
        self.assertEqual(m["type"], "claimed")
        self.assertEqual(m["mailbox"], mailbox_id) # mailbox id is stable
        np_row, side_rows = self._nameplate(app, "np1")
        claims = [(row["side"], row["claimed"]) for row in side_rows]
        self.assertEqual(claims, [("side", True)])
        c.close()
        yield c.d

        c = yield self.make_client()
        yield c.next_non_ack()
        c.send("bind", appid="appid", side="side")
        # we haven't done a claim with this particular connection, but we can
        # still send a release as long as we include the nameplate
        c.send("release", nameplate="np1") # release-without-claim
        m = yield c.next_non_ack()
        self.assertEqual(m["type"], "released")
        np_row, side_rows = self._nameplate(app, "np1")
        self.assertEqual(np_row, None)
        c.close()
        yield c.d

        c = yield self.make_client()
        yield c.next_non_ack()
        c.send("bind", appid="appid", side="side")
        # and the release is idempotent, when done on separate connections
        c.send("release", nameplate="np1")
        m = yield c.next_non_ack()
        self.assertEqual(m["type"], "released")
        np_row, side_rows = self._nameplate(app, "np1")
        self.assertEqual(np_row, None)
        c.close()
        yield c.d


    @inlineCallbacks
    def test_interrupted_client_nameplate_reclaimed(self):
        c = yield self.make_client()
        yield c.next_non_ack()
        c.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")

        # a new claim on a previously-closed nameplate is forbidden. We make
        # a new nameplate here and manually open a second claim on it, so the
        # nameplate stays alive long enough for the code check to happen.
        c = yield self.make_client()
        yield c.next_non_ack()
        c.send("bind", appid="appid", side="side")
        c.send("claim", nameplate="np2")
        m = yield c.next_non_ack()
        self.assertEqual(m["type"], "claimed")
        app.claim_nameplate("np2", "side2", 0)
        c.send("release", nameplate="np2")
        m = yield c.next_non_ack()
        self.assertEqual(m["type"], "released")
        np_row, side_rows = self._nameplate(app, "np2")
        claims = sorted([(row["side"], row["claimed"]) for row in side_rows])
        self.assertEqual(claims, [("side", 0), ("side2", 1)])
        c.close()
        yield c.d

        c = yield self.make_client()
        yield c.next_non_ack()
        c.send("bind", appid="appid", side="side")
        c.send("claim", nameplate="np2") # new claim is forbidden
        err = yield c.next_non_ack()
        self.assertEqual(err["type"], "error")
        self.assertEqual(err["error"], "reclaimed")

        np_row, side_rows = self._nameplate(app, "np2")
        claims = sorted([(row["side"], row["claimed"]) for row in side_rows])
        self.assertEqual(claims, [("side", 0), ("side2", 1)])
        c.close()
        yield c.d

    @inlineCallbacks
    def test_interrupted_client_mailbox(self):
        # a client's interactions with the server might be split over
        # multiple sequential WebSocket connections, e.g. when the server is
        # bounced and the client reconnects, or vice versa
        c = yield self.make_client()
        yield c.next_non_ack()
        c.send("bind", appid="appid", side="side")
        app = self._server.get_app("appid")
        mb1 = app.open_mailbox("mb1", "side2", 0)
        mb1.add_message(SidedMessage(side="side2", phase="phase",
                                     body="body", server_rx=0,
                                     msg_id="msgid"))

        c.send("open", mailbox="mb1")
        m = yield c.next_non_ack()
        self.assertEqual(m["type"], "message")
        self.assertEqual(m["body"], "body")
        self.assertTrue(mb1.has_listeners())
        c.close()
        yield c.d

        c = yield self.make_client()
        yield c.next_non_ack()
        c.send("bind", appid="appid", side="side")
        # open should be idempotent
        c.send("open", mailbox="mb1")
        m = yield c.next_non_ack()
        self.assertEqual(m["type"], "message")
        self.assertEqual(m["body"], "body")
        mb_row, side_rows = self._mailbox(app, "mb1")
        openeds = [(row["side"], row["opened"]) for row in side_rows]
        self.assertIn(("side", 1), openeds) # TODO: why 1, and not True?

        # close on the same connection as open is ok
        c.send("close", mailbox="mb1", mood="mood")
        m = yield c.next_non_ack()
        self.assertEqual(m["type"], "closed", m)
        mb_row, side_rows = self._mailbox(app, "mb1")
        openeds = [(row["side"], row["opened"]) for row in side_rows]
        self.assertIn(("side", 0), openeds)
        c.close()
        yield c.d

        # close (on a separate connection) is idempotent
        c = yield self.make_client()
        yield c.next_non_ack()
        c.send("bind", appid="appid", side="side")
        c.send("close", mailbox="mb1", mood="mood")
        m = yield c.next_non_ack()
        self.assertEqual(m["type"], "closed", m)
        mb_row, side_rows = self._mailbox(app, "mb1")
        openeds = [(row["side"], row["opened"]) for row in side_rows]
        self.assertIn(("side", 0), openeds)
        c.close()
        yield c.d


